import hub.helpers.constants as cte

BUILDING_SURFACE = '\nBUILDINGSURFACE:DETAILED,\n'
WINDOW_SURFACE = '\nFENESTRATIONSURFACE:DETAILED,\n'
COMPACT_SCHEDULE = '\nSCHEDULE:COMPACT,\n'
FILE_SCHEDULE = '\nSCHEDULE:FILE,\n'
NOMASS_MATERIAL = '\nMATERIAL:NOMASS,\n'
SOLID_MATERIAL = '\nMATERIAL,\n'
WINDOW_MATERIAL = '\nWINDOWMATERIAL:SIMPLEGLAZINGSYSTEM,\n'
CONSTRUCTION = '\nCONSTRUCTION,\n'
ZONE = '\nZONE,\n'
GLOBAL_GEOMETRY_RULES = '\nGlobalGeometryRules,\n'
PEOPLE = '\nPEOPLE,\n'
LIGHTS = '\nLIGHTS,\n'
APPLIANCES = '\nOTHEREQUIPMENT,\n'
OUTPUT_CONTROL = '\nOutputControl:IlluminanceMap:Style,\n'
INFILTRATION = '\nZONEINFILTRATION:DESIGNFLOWRATE,\n'
VENTILATION = '\nZONEVENTILATION:DESIGNFLOWRATE,\n'
THERMOSTAT = '\nHVACTEMPLATE:THERMOSTAT,\n'
IDEAL_LOAD_SYSTEM = '\nHVACTEMPLATE:ZONE:IDEALLOADSAIRSYSTEM,\n'
DHW = '\nWATERUSE:EQUIPMENT,\n'
SHADING = '\nSHADING:BUILDING:DETAILED,\n'

AUTOCALCULATE = 'autocalculate'
ROUGHNESS = 'MediumRough'
OUTDOORS = 'Outdoors'
GROUND = 'Ground'
SURFACE = 'Surface'
SUN_EXPOSED = 'SunExposed'
WIND_EXPOSED = 'WindExposed'
NON_SUN_EXPOSED = 'NoSun'
NON_WIND_EXPOSED = 'NoWind'
EMPTY = ''

idf_surfaces_dictionary = {
  cte.WALL: 'wall',
  cte.GROUND: 'floor',
  cte.ROOF: 'roof'
}

idf_type_limits = {
  cte.ON_OFF: 'on/off',
  cte.FRACTION: 'Fraction',
  cte.ANY_NUMBER: 'Any Number',
  cte.CONTINUOUS: 'Continuous',
  cte.DISCRETE: 'Discrete'
}

idf_day_types = {
  cte.MONDAY: 'Monday',
  cte.TUESDAY: 'Tuesday',
  cte.WEDNESDAY: 'Wednesday',
  cte.THURSDAY: 'Thursday',
  cte.FRIDAY: 'Friday',
  cte.SATURDAY: 'Saturday',
  cte.SUNDAY: 'Sunday',
  cte.HOLIDAY: 'Holidays',
  cte.WINTER_DESIGN_DAY: 'WinterDesignDay',
  cte.SUMMER_DESIGN_DAY: 'SummerDesignDay'
}


import hub.exports.building_energy.idf_helper as idf_cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfWindowsConstructions(IdfBase):
  @staticmethod
  def add(self, thermal_boundary):
    name = f'{thermal_boundary.construction_name}_window'
    if name not in self._windows_added_to_idf:
      return  # Material not added or already assigned to construction
    construction_name = f'{thermal_boundary.construction_name}_window_construction'
    if construction_name not in self._constructions_added_to_idf:
      self._constructions_added_to_idf[construction_name] = True
      file = self._files['constructions']
      self._write_to_idf_format(file, idf_cte.CONSTRUCTION)
      self._write_to_idf_format(file, construction_name, 'Name')
      self._write_to_idf_format(file, name, 'Outside Layer', ';')



import hub.exports.building_energy.idf_helper as idf_cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfShading(IdfBase):
  @staticmethod
  def add(self, building):
    name = building.name
    file = self._files['shading']
    for s, surface in enumerate(building.surfaces):

      self._write_to_idf_format(file, idf_cte.SHADING)
      self._write_to_idf_format(file, f'{name}_{s}', 'Name')
      self._write_to_idf_format(file, idf_cte.EMPTY, 'Transmittance Schedule Name')
      self._write_to_idf_format(file, idf_cte.AUTOCALCULATE, 'Number of Vertices')
      eol = ','
      coordinates = self._matrix_to_list(surface.solid_polygon.coordinates, self._city.lower_corner)
      coordinates_length = len(coordinates)
      for i, coordinate in enumerate(coordinates):
        vertex = i + 1
        if vertex == coordinates_length:
          eol = ';'
        self._write_to_idf_format(file, coordinate[0], f'Vertex {vertex} Xcoordinate')
        self._write_to_idf_format(file, coordinate[1], f'Vertex {vertex} Ycoordinate')
        self._write_to_idf_format(file, coordinate[2], f'Vertex {vertex} Zcoordinate', eol)



import hub.exports.building_energy.idf_helper as idf_cte
import hub.helpers.constants as cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfSurfaces(IdfBase):
  @staticmethod
  def add(self, building, zone_name):
    zone_name = f'{zone_name}'
    file = self._files['surfaces']
    for thermal_zone in building.thermal_zones_from_internal_zones:
      for index, boundary in enumerate(thermal_zone.thermal_boundaries):
        surface_type = idf_cte.idf_surfaces_dictionary[boundary.parent_surface.type]
        outside_boundary_condition = idf_cte.OUTDOORS
        sun_exposure = idf_cte.SUN_EXPOSED
        wind_exposure = idf_cte.WIND_EXPOSED
        outside_boundary_condition_object = idf_cte.EMPTY
        name = f'Building_{building.name}_surface_{index}'
        construction_name = f'{boundary.construction_name} {boundary.parent_surface.type}'
        space_name = idf_cte.EMPTY
        if boundary.parent_surface.type == cte.GROUND:
          outside_boundary_condition = idf_cte.GROUND
          sun_exposure = idf_cte.NON_SUN_EXPOSED
          wind_exposure = idf_cte.NON_WIND_EXPOSED
        if boundary.parent_surface.percentage_shared is not None and boundary.parent_surface.percentage_shared > 0.5:
          outside_boundary_condition_object = f'Building_{building.name}_surface_{index}'
          outside_boundary_condition = idf_cte.SURFACE
          sun_exposure = idf_cte.NON_SUN_EXPOSED
          wind_exposure = idf_cte.NON_WIND_EXPOSED
        self._write_to_idf_format(file, idf_cte.BUILDING_SURFACE)
        self._write_to_idf_format(file, name, 'Name')
        self._write_to_idf_format(file, surface_type, 'Surface Type')
        self._write_to_idf_format(file, construction_name, 'Construction Name')
        self._write_to_idf_format(file, zone_name, 'Zone Name')
        self._write_to_idf_format(file, space_name, 'Space Name')
        self._write_to_idf_format(file, outside_boundary_condition, 'Outside Boundary Condition')
        self._write_to_idf_format(file, outside_boundary_condition_object, 'Outside Boundary Condition Object')
        self._write_to_idf_format(file, sun_exposure, 'Sun Exposure')
        self._write_to_idf_format(file, wind_exposure, 'Wind Exposure')
        self._write_to_idf_format(file, idf_cte.AUTOCALCULATE, 'View Factor to Ground')
        self._write_to_idf_format(file, idf_cte.AUTOCALCULATE, 'Number of Vertices')
        coordinates = self._matrix_to_list(boundary.parent_surface.solid_polygon.coordinates,
                                           self._city.lower_corner)
        eol = ','
        coordinates_length = len(coordinates)
        for i, coordinate in enumerate(coordinates):
          vertex = i + 1
          if vertex == coordinates_length:
            eol = ';'
          self._write_to_idf_format(file, coordinate[0], f'Vertex {vertex} Xcoordinate')
          self._write_to_idf_format(file, coordinate[1], f'Vertex {vertex} Ycoordinate')
          self._write_to_idf_format(file, coordinate[2], f'Vertex {vertex} Zcoordinate', eol)



from pathlib import Path
import hub.helpers.constants as cte
import hub.exports.building_energy.idf_helper as idf_cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfFileSchedule(IdfBase):
  @staticmethod
  def add(self, usage, schedule_type, schedules):
    schedule_name = f'{schedule_type} schedules {usage}'
    for schedule in schedules:
      if schedule_name not in self._schedules_added_to_idf:
        self._schedules_added_to_idf[schedule_name] = True
        file_name = str(
          (Path(self._output_path) / f'{schedule_type} schedules {usage.replace("/", "_")}.csv').resolve())
        with open(file_name, 'w', encoding='utf8') as file:
          for value in schedule.values[0]:
            file.write(f'{value},\n')
        file = self._files['file_schedules']
        self._write_to_idf_format(file, idf_cte.FILE_SCHEDULE)
        self._write_to_idf_format(file, schedule_name, 'Name')
        self._write_to_idf_format(file, idf_cte.idf_type_limits[schedule.data_type], 'Schedule Type Limits Name')
        self._write_to_idf_format(file, Path(file_name).name, 'File Name')
        self._write_to_idf_format(file, 1, 'Column Number')
        self._write_to_idf_format(file, 0, 'Rows to Skip at Top')
        self._write_to_idf_format(file, 8760, 'Number of Hours of Data')
        self._write_to_idf_format(file, 'Comma', 'Column Separator')
        self._write_to_idf_format(file, 'No', 'Interpolate to Timestep')
        self._write_to_idf_format(file, '60', 'Minutes per Item')
        self._write_to_idf_format(file, 'Yes', 'Adjust Schedule for Daylight Savings', ';')



import hub.exports.building_energy.idf_helper as idf_cte
import hub.helpers.constants as cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfVentilation(IdfBase):
  @staticmethod
  def add(self, thermal_zone, zone_name):
    schedule_name = f'Ventilation schedules {thermal_zone.usage_name}'
    air_change = thermal_zone.mechanical_air_change * cte.HOUR_TO_SECONDS
    file = self._files['ventilation']
    self._write_to_idf_format(file, idf_cte.VENTILATION)
    self._write_to_idf_format(file, f'{zone_name}_ventilation', 'Name')
    self._write_to_idf_format(file, zone_name, 'Zone or ZoneList or Space or SpaceList Name')
    self._write_to_idf_format(file, schedule_name, 'Schedule Name')
    self._write_to_idf_format(file, 'AirChanges/Hour', 'Design Flow Rate Calculation Method')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Design Flow Rate')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Flow Rate per Floor Area')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Flow Rate per Person')
    self._write_to_idf_format(file, air_change, 'Air Changes per Hour')
    self._write_to_idf_format(file, 'Natural', 'Ventilation Type')
    self._write_to_idf_format(file, 0, 'Fan Pressure Rise')
    self._write_to_idf_format(file, 1, 'Fan Total Efficiency')
    self._write_to_idf_format(file, 1, 'Constant Term Coefficient')
    self._write_to_idf_format(file, 0, 'Temperature Term Coefficient')
    self._write_to_idf_format(file, 0, 'Velocity Term Coefficient')
    self._write_to_idf_format(file, 0, 'Velocity Squared Term Coefficient')
    self._write_to_idf_format(file, -100, 'Minimum Indoor Temperature')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Minimum Indoor Temperature Schedule Name')
    self._write_to_idf_format(file, 100, 'Maximum Indoor Temperature')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Maximum Indoor Temperature Schedule Name')
    self._write_to_idf_format(file, -100, 'Delta Temperature')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Delta Temperature Schedule Name')
    self._write_to_idf_format(file, -100, 'Minimum Outdoor Temperature')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Minimum Outdoor Temperature Schedule Name')
    self._write_to_idf_format(file, 100, 'Maximum Outdoor Temperature')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Maximum Outdoor Temperature Schedule Name')
    self._write_to_idf_format(file, 40, 'Maximum Wind Speed', ';')



import os
from pathlib import Path

import hub.exports.building_energy.idf_helper as idf_cte


class IdfBase:
  def __init__(self, city, output_path, idf_file_path, idd_file_path, epw_file_path, target_buildings=None,
               _calculate_with_new_infiltration=True):
    self._city = city
    self._output_path = str(output_path.resolve())
    self._output_file_path = str((output_path / f'{city.name}.idf').resolve())

    self._file_paths = {
      'schedules': str((output_path / 'schedules.idf').resolve()),
      'file_schedules': str((output_path / 'file_schedules.idf').resolve()),
      'solid_materials': str((output_path / 'solid_materials.idf').resolve()),
      'nomass_materials': str((output_path / 'nomass_materials.idf').resolve()),
      'window_materials': str((output_path / 'window_materials.idf').resolve()),
      'constructions': str((output_path / 'constructions.idf').resolve()),
      'zones': str((output_path / 'zones.idf').resolve()),
      'surfaces': str((output_path / 'surfaces.idf').resolve()),
      'fenestration': str((output_path / 'fenestration.idf').resolve()),
      'occupancy': str((output_path / 'occupancy.idf').resolve()),
      'lighting': str((output_path / 'lights.idf').resolve()),
      'appliances': str((output_path / 'appliances.idf').resolve()),
      'shading': str((output_path / 'shading.idf').resolve()),
      'infiltration': str((output_path / 'infiltration.idf').resolve()),
      'ventilation': str((output_path / 'ventilation.idf').resolve()),
      'thermostat': str((output_path / 'thermostat.idf').resolve()),
      'ideal_load_system': str((output_path / 'ideal_load_system.idf').resolve()),
      'dhw': str((output_path / 'dhw.idf').resolve()),
    }
    self._files = {}
    for key, value in self._file_paths.items():
      self._files[key] = open(value, 'w', encoding='UTF-8')

    self._idd_file_path = str(idd_file_path)
    self._idf_file_path = str(idf_file_path)
    self._outputs_file_path = str(Path(idf_file_path).parent / 'outputs.idf')
    self._epw_file_path = str(epw_file_path)

    self._target_buildings = target_buildings
    self._adjacent_buildings = []

    if target_buildings is None:
      self._target_buildings = [building.name for building in self._city.buildings]
    else:
      for building_name in target_buildings:
        building = city.city_object(building_name)
        if building.neighbours is not None:
          self._adjacent_buildings += building.neighbours
    self._calculate_with_new_infiltration = _calculate_with_new_infiltration

  def _create_output_control_lighting(self):
    file = self._files['appliances']
    self._write_to_idf_format(file, idf_cte.OUTPUT_CONTROL)
    self._write_to_idf_format(file, 'Comma', 'Column Separator', ';')

  @staticmethod
  def _write_to_idf_format(file, field, comment='', eol=','):
    if comment != '':
      comment = f'    !- {comment}'
      field = f'    {field}{eol}'.ljust(26, ' ')
      file.write(f'{field}{comment}\n')
    else:
      file.write(f'{field}{comment}')

  @staticmethod
  def _matrix_to_list(points, lower_corner):
    lower_x = lower_corner[0]
    lower_y = lower_corner[1]
    lower_z = lower_corner[2]
    points_list = []
    for point in points:
      point_tuple = (point[0] - lower_x, point[1] - lower_y, point[2] - lower_z)
      points_list.append(point_tuple)
    return points_list



import hub.exports.building_energy.idf_helper as idf_cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfHeatingSystem(IdfBase):
  @staticmethod
  def add(self, thermal_zone, zone_name):
    availability_schedule = f'HVAC AVAIL SCHEDULES {thermal_zone.usage_name}'
    thermostat_name = f'Thermostat {thermal_zone.usage_name}'
    file = self._files['ideal_load_system']
    self._write_to_idf_format(file, idf_cte.IDEAL_LOAD_SYSTEM)
    self._write_to_idf_format(file, zone_name, 'Zone Name')
    self._write_to_idf_format(file, thermostat_name, 'Template Thermostat Name')
    self._write_to_idf_format(file, availability_schedule, 'System Availability Schedule Name')
    self._write_to_idf_format(file, 50, 'Maximum Heating Supply Air Temperature')
    self._write_to_idf_format(file, 13, 'Minimum Cooling Supply Air Temperature')
    self._write_to_idf_format(file, 0.0156, 'Maximum Heating Supply Air Humidity Ratio')
    self._write_to_idf_format(file, 0.0077, 'Minimum Cooling Supply Air Humidity Ratio')
    self._write_to_idf_format(file, 'NoLimit', 'Heating Limit')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Maximum Heating Air Flow Rate')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Maximum Sensible Heating Capacity')
    self._write_to_idf_format(file, 'NoLimit', 'Cooling Limit')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Maximum Cooling Air Flow Rate')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Maximum Total Cooling Capacity')
    self._write_to_idf_format(file, availability_schedule, 'Heating Availability Schedule Name')
    self._write_to_idf_format(file, availability_schedule, 'Cooling Availability Schedule Name')
    self._write_to_idf_format(file, 'ConstantSensibleHeatRatio', 'Dehumidification Control Type')
    self._write_to_idf_format(file, 0.7, 'Cooling Sensible Heat Ratio')
    self._write_to_idf_format(file, 60, 'Dehumidification Setpoint')
    self._write_to_idf_format(file, 'None', 'Humidification Control Type')
    self._write_to_idf_format(file, 30, 'Humidification Setpoint')
    self._write_to_idf_format(file, 'None', 'Outdoor Air Method')
    self._write_to_idf_format(file, 0.00944, 'Outdoor Air Flow Rate per Person')
    self._write_to_idf_format(file, 0.0, 'Outdoor Air Flow Rate per Zone Floor Area')
    self._write_to_idf_format(file, 0, 'Outdoor Air Flow Rate per Zone')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Design Specification Outdoor Air Object Name')
    self._write_to_idf_format(file, 'None', 'Demand Controlled Ventilation Type')
    self._write_to_idf_format(file, 'NoEconomizer', 'Outdoor Air Economizer Type')
    self._write_to_idf_format(file, 'None', 'Heat Recovery Type')
    self._write_to_idf_format(file, 0.70, 'Sensible Heat Recovery Effectiveness')
    self._write_to_idf_format(file, 0.65, 'Latent Heat Recovery Effectiveness', ';')


import hub.exports.building_energy.idf_helper as idf_cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfDhw(IdfBase):
  @staticmethod
  def add(self, thermal_zone, zone_name):
    peak_flow_rate = thermal_zone.domestic_hot_water.peak_flow * thermal_zone.total_floor_area
    flow_rate_schedule = f'DHW_prof schedules {thermal_zone.usage_name}'
    dhw_schedule = f'DHW_temp schedules {thermal_zone.usage_name}'
    cold_temp_schedule = f'cold_temp schedules {thermal_zone.usage_name}'
    file = self._files['dhw']
    self._write_to_idf_format(file, idf_cte.DHW)
    self._write_to_idf_format(file, zone_name, 'Name')
    self._write_to_idf_format(file, zone_name, 'EndUse Subcategory')
    self._write_to_idf_format(file, peak_flow_rate, 'Peak Flow Rate')
    self._write_to_idf_format(file, flow_rate_schedule, 'Flow Rate Fraction Schedule Name')
    self._write_to_idf_format(file, dhw_schedule, 'Target Temperature Schedule Name')
    self._write_to_idf_format(file, dhw_schedule, 'Hot Water Supply Temperature Schedule Name')
    self._write_to_idf_format(file, cold_temp_schedule, 'Cold Water Supply Temperature Schedule Name')
    self._write_to_idf_format(file, zone_name, 'Zone Name', ';')



import hub.exports.building_energy.idf_helper as idf_cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfSchedule(IdfBase):
  @staticmethod
  def add(self, usage, schedule_type, schedules):
    if len(schedules) < 1:
      return
    schedule_name = f'{schedule_type} schedules {usage}'
    if schedule_name not in self._schedules_added_to_idf:
      self._schedules_added_to_idf[schedule_name] = True
      file = self._files['schedules']
      self._write_to_idf_format(file, idf_cte.COMPACT_SCHEDULE)
      self._write_to_idf_format(file, schedule_name, 'Name')
      self._write_to_idf_format(file, idf_cte.idf_type_limits[schedules[0].data_type], 'Schedule Type Limits Name')
      self._write_to_idf_format(file, 'Through: 12/31', 'Field 1')
      counter = 1
      for j, schedule in enumerate(schedules):
        _val = schedule.values
        _new_field = ''
        for day_type in schedule.day_types:
          _new_field += f' {idf_cte.idf_day_types[day_type]}'
        self._write_to_idf_format(file, f'For:{_new_field}', f'Field {j * 25 + 2}')
        counter += 1
        for i, _ in enumerate(_val):
          self._write_to_idf_format(file, f'Until: {i + 1:02d}:00,{_val[i]}', f'Field {j * 25 + 3 + i}')
          counter += 1
      self._write_to_idf_format(file, 'For AllOtherDays', f'Field {counter + 1}')
      self._write_to_idf_format(file, 'Until: 24:00,0.0', f'Field {counter + 2}', ';')



import hub.exports.building_energy.idf_helper as idf_cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfZone(IdfBase):
  @staticmethod
  def add(self, thermal_zone, zone_name):
    file = self._files['zones']
    self._write_to_idf_format(file, idf_cte.ZONE)
    self._write_to_idf_format(file, zone_name, 'Name')
    self._write_to_idf_format(file, 0, 'Direction of Relative North')
    self._write_to_idf_format(file, 0, 'X Origin')
    self._write_to_idf_format(file, 0, 'Y Origin')
    self._write_to_idf_format(file, 0, 'Z Origin')
    self._write_to_idf_format(file, 1, 'Type')
    self._write_to_idf_format(file, 1, 'Multiplier')
    self._write_to_idf_format(file, idf_cte.AUTOCALCULATE, 'Ceiling Height')
    self._write_to_idf_format(file, thermal_zone.volume, 'Volume')
    self._write_to_idf_format(file, idf_cte.AUTOCALCULATE, 'Floor Area')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Zone Inside Convection Algorithm')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Zone Outside Convection Algorithm')
    self._write_to_idf_format(file, 'Yes', 'Part of Total Floor Area', ';')


import hub.exports.building_energy.idf_helper as idf_cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfAppliance(IdfBase):
  @staticmethod
  def add(self, thermal_zone, zone_name):
    schedule_name = f'Appliance schedules {thermal_zone.usage_name}'
    storeys_number = int(thermal_zone.total_floor_area / thermal_zone.footprint_area)
    watts_per_zone_floor_area = thermal_zone.appliances.density * storeys_number
    subcategory = f'ELECTRIC EQUIPMENT#{zone_name}#InteriorEquipment'
    file = self._files['appliances']
    self._write_to_idf_format(file, idf_cte.APPLIANCES)
    self._write_to_idf_format(file, zone_name, 'Name')
    self._write_to_idf_format(file, 'Electricity', 'Fuel Type')
    self._write_to_idf_format(file, zone_name, 'Zone or ZoneList or Space or SpaceList Name')
    self._write_to_idf_format(file, schedule_name, 'Schedule Name')
    self._write_to_idf_format(file, 'Watts/Area', 'Design Level Calculation Method')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Design Level')
    self._write_to_idf_format(file, watts_per_zone_floor_area, 'Power per Zone Floor Area')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Power per Person')
    self._write_to_idf_format(file, thermal_zone.appliances.latent_fraction, 'Fraction Latent')
    self._write_to_idf_format(file, thermal_zone.appliances.radiative_fraction, 'Fraction Radiant')
    self._write_to_idf_format(file, 0, 'Fraction Lost')
    self._write_to_idf_format(file, 0, 'Carbon Dioxide Generation Rate')
    self._write_to_idf_format(file, subcategory, 'EndUse Subcategory', ';')



import hub.exports.building_energy.idf_helper as idf_cte
from hub.city_model_structure.building_demand.layer import Layer
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfConstruction(IdfBase):

  @staticmethod
  def _add_solid_material(self, layer):
    file = self._files['solid_materials']
    self._write_to_idf_format(file, idf_cte.SOLID_MATERIAL)
    self._write_to_idf_format(file, layer.material_name, 'Name')
    self._write_to_idf_format(file, idf_cte.ROUGHNESS, 'Roughness')
    self._write_to_idf_format(file, layer.thickness, 'Thickness')
    self._write_to_idf_format(file, layer.conductivity, 'Conductivity')
    self._write_to_idf_format(file, layer.density, 'Density')
    self._write_to_idf_format(file, layer.specific_heat, 'Specific Heat')
    self._write_to_idf_format(file, layer.thermal_absorptance, 'Thermal Absorptance')
    self._write_to_idf_format(file, layer.solar_absorptance, 'Solar Absorptance')
    self._write_to_idf_format(file, layer.visible_absorptance, 'Visible Absorptance', ';')

  @staticmethod
  def _add_default_material(self):
    layer = Layer()
    layer.material_name = 'DefaultMaterial'
    layer.thickness = 0.1
    layer.conductivity = 0.1
    layer.density = 1000
    layer.specific_heat = 1000
    layer.thermal_absorptance = 0.9
    layer.solar_absorptance = 0.9
    layer.visible_absorptance = 0.7
    IdfConstruction._add_solid_material(self, layer)
    return layer

  @staticmethod
  def add(self, thermal_boundary):
    if thermal_boundary.layers is None:
      thermal_boundary.layers = [IdfConstruction._add_default_material(self)]
    name = f'{thermal_boundary.construction_name} {thermal_boundary.parent_surface.type}'

    if name not in self._constructions_added_to_idf:
      self._constructions_added_to_idf[name] = True
      file = self._files['constructions']
      self._write_to_idf_format(file, idf_cte.CONSTRUCTION)
      self._write_to_idf_format(file, name, 'Name')
      eol = ','
      if len(thermal_boundary.layers) == 1:
        eol = ';'
      self._write_to_idf_format(file, thermal_boundary.layers[0].material_name, 'Outside Layer', eol)
      for i in range(1, len(thermal_boundary.layers) - 1):
        comment = f'Layer {i + 1}'
        material_name = thermal_boundary.layers[i].material_name
        if i == len(thermal_boundary.layers) - 2:
          eol = ';'
        self._write_to_idf_format(file, material_name, comment, eol)



import hub.exports.building_energy.idf_helper as idf_cte
import hub.helpers.constants as cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfInfiltration(IdfBase):
  @staticmethod
  def add(self, thermal_zone, zone_name):
    IdfInfiltration._add_infiltration(self, thermal_zone, zone_name, 'AirChanges/Hour', cte.HOUR_TO_SECONDS)

  @staticmethod
  def add_surface(self, thermal_zone, zone_name):
    IdfInfiltration._add_infiltration(self, thermal_zone, zone_name, 'Flow/ExteriorWallArea', cte.INFILTRATION_75PA_TO_4PA)

  @staticmethod
  def _add_infiltration(self, thermal_zone, zone_name, calculation_method, multiplier):
    schedule_name = f'Infiltration schedules {thermal_zone.usage_name}'
    infiltration = thermal_zone.infiltration_rate_system_off * multiplier
    file = self._files['infiltration']
    self._write_to_idf_format(file, idf_cte.INFILTRATION)
    self._write_to_idf_format(file, zone_name, 'Name')
    self._write_to_idf_format(file, zone_name, 'Zone or ZoneList or Space or SpaceList Name')
    self._write_to_idf_format(file, schedule_name, 'Schedule Name')
    self._write_to_idf_format(file, calculation_method, 'Design Flow Rate Calculation Method')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Design Flow Rate')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Flow Rate per Floor Area')
    self._write_to_idf_format(file, infiltration, 'Flow Rate per Exterior Surface Area')
    self._write_to_idf_format(file, infiltration, 'Air Changes per Hour')
    self._write_to_idf_format(file, 1, 'Constant Term Coefficient')
    self._write_to_idf_format(file, 0, 'Temperature Term Coefficient')
    self._write_to_idf_format(file, 0, 'Velocity Term Coefficient')
    self._write_to_idf_format(file, 0, 'Velocity Squared Term Coefficient', ';')



import hub.exports.building_energy.idf_helper as idf_cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfWindowsMaterial(IdfBase):
  @staticmethod
  def add(self, thermal_boundary, thermal_opening):
    name = f'{thermal_boundary.construction_name}_window'
    if name not in self._windows_added_to_idf:
      self._windows_added_to_idf[name] = True
      file = self._files['window_materials']
      self._write_to_idf_format(file, idf_cte.WINDOW_MATERIAL)
      self._write_to_idf_format(file, name, 'Name')
      self._write_to_idf_format(file, thermal_opening.overall_u_value, 'UFactor')
      self._write_to_idf_format(file, thermal_opening.g_value, 'Solar Heat Gain Coefficient', ';')



import logging

import hub.exports.building_energy.idf_helper as idf_cte
import hub.helpers.constants as cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfWindow(IdfBase):

  @staticmethod
  def _to_window_surface(self, surface):
    window_ratio = surface.associated_thermal_boundaries[0].window_ratio
    x = 0
    y = 1
    z = 2
    coordinates = self._matrix_to_list(surface.solid_polygon.coordinates, self._city.lower_corner)
    min_z = surface.lower_corner[z]
    max_z = surface.upper_corner[z]
    middle = (max_z - min_z) / 2
    distance = (max_z - min_z) * window_ratio
    new_max_z = middle + distance / 2
    new_min_z = middle - distance / 2
    for index, coordinate in enumerate(coordinates):
      if coordinate[z] == max_z:
        coordinates[index] = (coordinate[x], coordinate[y], new_max_z)
      elif coordinate[z] == min_z:
        coordinates[index] = (coordinate[x], coordinate[y], new_min_z)
      else:
         logging.warning('Z coordinate not in top or bottom during window creation')
    return coordinates

  @staticmethod
  def add(self, building):
    file = self._files['fenestration']
    for thermal_zone in building.thermal_zones_from_internal_zones:
      for index, boundary in enumerate(thermal_zone.thermal_boundaries):
        building_surface_name = f'Building_{building.name}_surface_{index}'
        is_exposed = boundary.parent_surface.type == cte.WALL
        if boundary.parent_surface.percentage_shared is not None and boundary.parent_surface.percentage_shared > 0.5 or boundary.window_ratio == 0:
          is_exposed = False
        if not is_exposed:
          continue
        name = f'Building_{building.name}_window_{index}'
        construction_name = f'{boundary.construction_name}_window_construction'
        self._write_to_idf_format(file, idf_cte.WINDOW_SURFACE)
        self._write_to_idf_format(file, name, 'Name')
        self._write_to_idf_format(file, 'Window', 'Surface Type')
        self._write_to_idf_format(file, construction_name, 'Construction Name')
        self._write_to_idf_format(file, building_surface_name, 'Building Surface Name')
        self._write_to_idf_format(file, idf_cte.EMPTY, 'Outside Boundary Condition Object')
        self._write_to_idf_format(file, idf_cte.AUTOCALCULATE, 'View Factor to Ground')
        self._write_to_idf_format(file, idf_cte.EMPTY, 'Frame and Divider Name')
        self._write_to_idf_format(file, '1.0', 'Multiplier')
        self._write_to_idf_format(file, idf_cte.AUTOCALCULATE, 'Number of Vertices')
        coordinates = IdfWindow._to_window_surface(self, boundary.parent_surface)
        eol = ','
        coordinates_length = len(coordinates)
        for i, coordinate in enumerate(coordinates):
          vertex = i + 1
          if vertex == coordinates_length:
            eol = ';'
          self._write_to_idf_format(file, coordinate[0], f'Vertex {vertex} Xcoordinate')
          self._write_to_idf_format(file, coordinate[1], f'Vertex {vertex} Ycoordinate')
          self._write_to_idf_format(file, coordinate[2], f'Vertex {vertex} Zcoordinate', eol)



import hub.exports.building_energy.idf_helper as idf_cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfMaterial(IdfBase):
  @staticmethod
  def _add_solid_material(self, layer):
    file = self._files['solid_materials']
    self._write_to_idf_format(file, idf_cte.SOLID_MATERIAL)
    self._write_to_idf_format(file, layer.material_name, 'Name')
    self._write_to_idf_format(file, idf_cte.ROUGHNESS, 'Roughness')
    self._write_to_idf_format(file, layer.thickness, 'Thickness')
    self._write_to_idf_format(file, layer.conductivity, 'Conductivity')
    self._write_to_idf_format(file, layer.density, 'Density')
    self._write_to_idf_format(file, layer.specific_heat, 'Specific Heat')
    self._write_to_idf_format(file, layer.thermal_absorptance, 'Thermal Absorptance')
    self._write_to_idf_format(file, layer.solar_absorptance, 'Solar Absorptance')
    self._write_to_idf_format(file, layer.visible_absorptance, 'Visible Absorptance', ';')

  @staticmethod
  def _add_nomass_material(self, layer):
    file = self._files['nomass_materials']
    self._write_to_idf_format(file, idf_cte.NOMASS_MATERIAL)
    self._write_to_idf_format(file, layer.material_name, 'Name')
    self._write_to_idf_format(file, idf_cte.ROUGHNESS, 'Roughness')
    self._write_to_idf_format(file, layer.thermal_resistance, 'Thermal Resistance')
    self._write_to_idf_format(file, 0.9, 'Thermal Absorptance')
    self._write_to_idf_format(file, 0.7, 'Solar Absorptance')
    self._write_to_idf_format(file, 0.7, 'Visible Absorptance', ';')

  @staticmethod
  def add(self, thermal_boundary):
    for layer in thermal_boundary.layers:
      if layer.material_name not in self._materials_added_to_idf:
        self._materials_added_to_idf[layer.material_name] = True
        if layer.no_mass:
          IdfMaterial._add_nomass_material(self, layer)
        else:
          IdfMaterial._add_solid_material(self, layer)



import hub.exports.building_energy.idf_helper as idf_cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfThermostat(IdfBase):
  @staticmethod
  def add(self, thermal_zone):
    thermostat_name = f'Thermostat {thermal_zone.usage_name}'
    heating_schedule = f'Heating thermostat schedules {thermal_zone.usage_name}'
    cooling_schedule = f'Cooling thermostat schedules {thermal_zone.usage_name}'
    if thermostat_name not in self._thermostat_added_to_idf:
      self._thermostat_added_to_idf[thermostat_name] = True
      file = self._files['thermostat']
      self._write_to_idf_format(file, idf_cte.THERMOSTAT)
      self._write_to_idf_format(file, thermostat_name, 'Name')
      self._write_to_idf_format(file, heating_schedule, 'Heating Setpoint Schedule Name')
      self._write_to_idf_format(file, idf_cte.EMPTY, 'Constant Heating Setpoint')
      self._write_to_idf_format(file, cooling_schedule, 'Cooling Setpoint Schedule Name', ';')



import hub.exports.building_energy.idf_helper as idf_cte
from hub.exports.building_energy.idf_helper.idf_base import IdfBase


class IdfOccupancy(IdfBase):
  @staticmethod
  def add(self, thermal_zone, zone_name):
    number_of_people = thermal_zone.occupancy.occupancy_density * thermal_zone.total_floor_area
    fraction_radiant = 0
    total_sensible = (
      thermal_zone.occupancy.sensible_radiative_internal_gain + thermal_zone.occupancy.sensible_convective_internal_gain
    )
    if total_sensible != 0:
      fraction_radiant = thermal_zone.occupancy.sensible_radiative_internal_gain / total_sensible
    occupancy_schedule = f'Occupancy schedules {thermal_zone.usage_name}'
    activity_level_schedule = f'Activity Level schedules {thermal_zone.usage_name}'
    file = self._files['occupancy']
    self._write_to_idf_format(file, idf_cte.PEOPLE)
    self._write_to_idf_format(file, f'{zone_name}_occupancy', 'Name')
    self._write_to_idf_format(file, zone_name, 'Zone or ZoneList or Space or SpaceList Name')
    self._write_to_idf_format(file, occupancy_schedule, 'Number of People Schedule Name')
    self._write_to_idf_format(file, 'People', 'Number of People Calculation Method')
    self._write_to_idf_format(file, number_of_people, 'Number of People')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'People per Floor Area')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Floor Area per Person')
    self._write_to_idf_format(file, fraction_radiant, 'Fraction Radiant')
    self._write_to_idf_format(file, idf_cte.AUTOCALCULATE, 'Sensible Heat Fraction')
    self._write_to_idf_format(file, activity_level_schedule, 'Activity Level Schedule Name')
    self._write_to_idf_format(file, '3.82e-08', 'Carbon Dioxide Generation Rate')
    self._write_to_idf_format(file, 'No', 'Enable ASHRAE 55 Comfort Warnings')
    self._write_to_idf_format(file, 'EnclosureAveraged', 'Mean Radiant Temperature Calculation Type')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Surface NameAngle Factor List Name')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Work Efficiency Schedule Name')
    self._write_to_idf_format(file, 'ClothingInsulationSchedule', 'Clothing Insulation Calculation Method')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Clothing Insulation Calculation Method Schedule Name')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Clothing Insulation Schedule Name')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Air Velocity Schedule Name')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Thermal Comfort Model 1 Type')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Thermal Comfort Model 2 Type')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Thermal Comfort Model 3 Type')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Thermal Comfort Model 4 Type')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Thermal Comfort Model 5 Type')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Thermal Comfort Model 6 Type')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Thermal Comfort Model 7 Type')
    self._write_to_idf_format(file, idf_cte.EMPTY, 'Ankle Level Air Velocity Schedule Name')
    self._write_to_idf_format(file, '15.56', 'Cold Stress Temperature Threshold')
    self._write_to_idf_format(file, '30', 'Heat Stress Temperature Threshold', ';')
